/* SYNRF: Reserved names and other constants for reference concrete syntax.
*/
#include "config.h"
#include "entity.h"           /* Templates for entity control blocks. */
#include "synxtrn.h"          /* Declarations for concrete syntax constants. */
#include "adl.h"              /* Definitions for attribute list processing. */
UNCH key[NKEYS][REFNAMELEN+1] = {
     "ANY",
     "ATTLIST",
     "CDATA",
     "CONREF",
     "CURRENT",
     "DEFAULT",
     "DOCTYPE",
     "ELEMENT",
     "EMPTY",
     "ENDTAG",
     "ENTITIES",
     "ENTITY",
     "FIXED",
     "ID",
     "IDLINK",
     "IDREF",
     "IDREFS",
     "IGNORE",
     "IMPLIED",
     "INCLUDE",
     "INITIAL",
     "LINK",
     "LINKTYPE",
     "MD",
     "MS",
     "NAME",
     "NAMES",
     "NDATA",
     "NMTOKEN",
     "NMTOKENS",
     "NOTATION",
     "NUMBER",
     "NUMBERS",
     "NUTOKEN",
     "NUTOKENS",
     "O",
     "PCDATA",
     "PI",
     "POSTLINK",
     "PUBLIC",
     "RCDATA",
     "RE",
     "REQUIRED",
     "RESTORE",
     "RS",
     "SDATA",
     "SHORTREF",
     "SIMPLE",
     "SPACE",
     "STARTTAG",
     "SUBDOC",
     "SYSTEM",
     "TEMP",
     "USELINK",
     "USEMAP"
};
/*
Local Variables:
c-indent-level: 5
c-continued-statement-offset: 5
c-brace-offset: -5
c-argdecl-indent: 0
c-label-offset: -5
End:
*/
